﻿/**
 * @file cpld_protocol.c
 * @brief Library for communication with CPLD
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpld_protocol.h"
#include "fuse_relay_board_pinout.h"
#include "usart_lib.h"
#include "power_monitor.h"
#include "mgmt_protocol_master.h"
#include "errors.h"
#include <asf.h>
#include <ioport.h>
#include <util/delay.h>
#include <stdbool.h>
#include <stdio.h>

uint8_t LED_counter = 0;
bool error_LED = false, LED_changed = false;
struct spi_device spi_device_conf = {
	.id = SPI_SS
};

void setAddress( uint8_t address, uint8_t half )
{
		//set address
	ioport_set_value(CPLD_ADDR1, (address & 0x01) ? true : false);
	ioport_set_value(CPLD_ADDR2, (address & 0x02) ? true : false);
	ioport_set_value(CPLD_ADDR3, (address & 0x04) ? true : false);
	if (!half)
		ioport_set_pin_low(CPLD_ADDR0);
	else
		ioport_set_pin_high(CPLD_ADDR0);
}


/**
 * \brief Read 4 bits of a CPLD register
 * 
 * \param address Register address
 * \param half 0 - bits 0..4; 1 - bits 5..8
 * 
 * \return uint8_t
 */
uint8_t readHalfRegister( uint8_t address, uint8_t half )
{
	ioport_set_pin_low(CPLD_CLOCK);
	
	// configure IO pins as input
	ioport_configure_pin(CPLD_DATA0, IOPORT_DIR_INPUT);
	ioport_configure_pin(CPLD_DATA1, IOPORT_DIR_INPUT);
	ioport_configure_pin(CPLD_DATA2, IOPORT_DIR_INPUT);
	ioport_configure_pin(CPLD_DATA3, IOPORT_DIR_INPUT);
	
	// activate CPLD bus drivers
	ioport_set_pin_high(CPLD_READ);
	
	setAddress(address, half);


	asm volatile("nop");
	ioport_set_pin_high(CPLD_CLOCK);
	//set read to low
	asm volatile("nop");

	ioport_set_pin_low(CPLD_CLOCK);
	asm volatile("nop");

	//read data
	uint8_t value = (ioport_get_pin_level(CPLD_DATA0) ? 1 : 0) << 0;
	value |= (ioport_get_pin_level(CPLD_DATA1) ? 1 : 0) << 1;
	value |= (ioport_get_pin_level(CPLD_DATA2) ? 1 : 0) << 2;
	value |= (ioport_get_pin_level(CPLD_DATA3) ? 1 : 0) << 3;

	return value;
}

void writeHalfRegister( uint8_t address, uint8_t half, uint8_t value )
{
	ioport_set_pin_low(CPLD_CLOCK);
	
	// deactivate CPLD bus drivers
	ioport_set_pin_low(CPLD_READ);
	
	// configure IO pins as input
	ioport_configure_pin(CPLD_DATA0, IOPORT_DIR_OUTPUT);
	ioport_configure_pin(CPLD_DATA1, IOPORT_DIR_OUTPUT);
	ioport_configure_pin(CPLD_DATA2, IOPORT_DIR_OUTPUT);
	ioport_configure_pin(CPLD_DATA3, IOPORT_DIR_OUTPUT);
	
	setAddress(address, half);
	
	//set data to be written
	ioport_set_value(CPLD_DATA0, (value & 0x01) ? true : false);
	ioport_set_value(CPLD_DATA1, (value & 0x02) ? true : false);
	ioport_set_value(CPLD_DATA2, (value & 0x04) ? true : false);
	ioport_set_value(CPLD_DATA3, (value & 0x08) ? true : false);

	ioport_set_pin_high(CPLD_CLOCK);
	asm volatile("nop");
	ioport_set_pin_low(CPLD_CLOCK);
	
	//set write to low
	//ioport_set_pin_high(CPLD_READ); -- commented out mfenner
}

/**
 * \brief Read data from CPLD register
 * 
 * \param address Register Address
 * 
 * \return uint8_t Data from the register
 */
uint8_t readRegister( uint8_t address )
{
	//return (readHalfRegister(address, 1) << 4 | readHalfRegister(address, 0));
	#ifdef DBG_LEVEL_VERBOSE
	//printf("Reading from CPLD\r\n");
	#endif
	
	uint8_t data = 0;
	int8_t status = 0;
	address |= CPLD_SPI_READ_OFFSET;
	spi_select_device(&SPIE, &spi_device_conf);

	status = spi_write_packet(&SPIE, &address, 1);
	increment_spi_cpld(status);
	status = spi_read_packet(&SPIE, &data, 1);
	increment_spi_cpld(status);

	spi_deselect_device(&SPIE, &spi_device_conf);
	return data;
}

/**
 * \brief Write data to CPLD register
 * 
 * \param address Register Address
 * \param value Data to be written
 * 
 * \return void
 */
void writeRegister( uint8_t address, uint8_t value )
{
	//writeHalfRegister(address, 1, (value & 0xF0) >> 4);
	//writeHalfRegister(address, 0, (value & 0x0F));
	
	#ifdef DBG_LEVEL_VERBOSE
	//printf("Writing to CPLD\r\n");
	#endif
	
	uint8_t data[2] = {address, value};
	int8_t status = 0;
	spi_select_device(&SPIE, &spi_device_conf);

	status = spi_write_packet(&SPIE, data, 2);
	increment_spi_cpld(status);
	spi_deselect_device(&SPIE, &spi_device_conf);
}

uint8_t getCPLDversion( void )
{
	return readRegister(CPLD_ADDR_VERSION);
}

void setRelaysState( uint8_t state )
{
	
    uint8_t i=0;
    uint8_t RelayStatus=readRegister(CPLD_ADDR_RELAY_STATUS);
    // Find the Channels that shall be switched on;
    uint8_t turnon=state & ~RelayStatus;
    
    if (turnon)
    {
        // Mark Relays to do PWM with
        writeRegister(CPLD_ADDR_PWMVALUE, 0);
        // Turn PWM Selected channels on
        writeRegister(CPLD_ADDR_PWMENABLE, turnon);
        // Write New State to CPLD (PWM controlled channels will stay off)
        writeRegister(CPLD_ADDR_RELAY_CONTROL, state);
        // Count PWM up;
        for (i=0; i<255; i++) 
        {
           writeRegister(CPLD_ADDR_PWMVALUE, i);
           _delay_us(500);             
        }    
        writeRegister(CPLD_ADDR_PWMENABLE, 0);
        writeRegister(CPLD_ADDR_PWMVALUE, 0);
    }
	
    // Turn Channels permanently on/off
    writeRegister(CPLD_ADDR_RELAY_CONTROL, state);
}

uint8_t getRelaysState( void )
{
	return readRegister(CPLD_ADDR_RELAY_STATUS);
}

uint8_t getCPLDscratchpad( void )
{
	return readRegister(CPLD_ADDR_SCRATCHPAD);
}

void setCPLDscratchpad( uint8_t value )
{
	writeRegister(CPLD_ADDR_SCRATCHPAD, value);
}


uint8_t getAlarmPin0Mask( void )
{
	return readRegister(CPLD_ADDR_ALARMPIN0_MASK);
}

uint8_t getAlarmPin1Mask( void )
{
	return readRegister(CPLD_ADDR_ALARMPIN1_MASK);
}

void setAlarmPin0Mask( uint8_t value )
{
	writeRegister(CPLD_ADDR_ALARMPIN0_MASK, value);
}

void setAlarmPin1Mask( uint8_t value )
{
	writeRegister(CPLD_ADDR_ALARMPIN1_MASK, value);
}

uint8_t getAlarmPinsTriggered( void )
{
	return readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED);
}

void setTMCBPresence( uint8_t state )
{
	writeRegister(CPLD_ADDR_CONTROL, state);
}

uint8_t getTMCBPresence( void )
{
	return readRegister(CPLD_ADDR_CONTROL);
}

void CPLD_init( void )
{
	spi_master_init(&SPIE);
	spi_master_setup_device(&SPIE, &spi_device_conf, SPI_MODE_0, CPLD_SPI_SPEED, 0);
	spi_enable(&SPIE);
	enableTmcbMP();
}

void clearCpldInterrupt( void )
{
	uint8_t data = readRegister(CPLD_ADDR_CONTROL);
	writeRegister(CPLD_ADDR_CONTROL, data | CPLD_CONTROL_CLEAR_INT);
	writeRegister(CPLD_ADDR_CONTROL, data & ~CPLD_CONTROL_CLEAR_INT);
}

void setCpldErrorLED( bool on )
{
	if (on != error_LED)
	{
		error_LED = on;
		LED_changed = true;
	}
}

void refreshCpldErrorLED( void )
{
	if (LED_changed)
	{
		if (tmcbIsConfigured())
		{
			uint8_t data = readRegister(CPLD_ADDR_CONTROL);
			writeRegister(CPLD_ADDR_CONTROL,
				(error_LED) ? (data | CPLD_CONTROL_ERROR_LED)
				: (data & ~CPLD_CONTROL_ERROR_LED));
			LED_counter = (error_LED) ? CPLD_ERROR_LED_CLEAR_TIME_S*10 : 0;
		}
	}
	LED_changed = false;
	if (--LED_counter == 0)
		setCpldErrorLED(false);
}

void enableTmcbMP( void )
{
	uint8_t data = readRegister(CPLD_ADDR_CONTROL);
	writeRegister(CPLD_ADDR_CONTROL, data | CPLD_CONTROL_ENABLE_TMCB_MP);
}

void disableTmcbMP( void )
{
	uint8_t data = readRegister(CPLD_ADDR_CONTROL);
	writeRegister(CPLD_ADDR_CONTROL, data & ~CPLD_CONTROL_ENABLE_TMCB_MP);
}

void clearAlarms( void )
{
	writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 0x00);
	for (int i = 0; i < CHANNEL_COUNT; ++i)
	{
		channels[i].alarm_pin0_status = false;
		channels[i].alarm_pin1_status = false;
		clearOvercurrentAlarm(i);
	}
}
