﻿/**
* @file fan_lib.c
* @brief Test of communication between CPLD and uC
*
* @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
* @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
* @author MICHAEL FENNER (DESY)
* @par License
*
* Redistribution and use in source and binary forms, with or without
* modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
*
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include "fan_lib.h"
#include "usart_lib.h"
#include "power_monitor.h"
#include "telnet_protocol.h"
#include "cpld_protocol.h"
#include "mgmt_protocol_master.h"
#include "fan_current_monitor_lib.h"
#include "i2c_lib.h"
#include "eeprom.h"
#include <asf.h>
#include <util/delay.h>
#include <avr/eeprom.h>


unsigned fan_target_rpm[FAN_COUNT] = {0};
uint8_t forceFanOff = 1;

int fan_CheckPresence(void)
{
	uint8_t data, reg;
	reg=FANREG_CONFIG;
	if ( i2c_read(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1) == TWI_SUCCESS) return STATUS_OK;
	else
	return ERR_IO_ERROR;
}

void fan_init(void)
{
	//i2c_init(&EXTERNAL_I2C);
	printPString("\r\n%% Initializing FAN control IC \r\n");
	if (fan_CheckPresence()==ERR_IO_ERROR)
	//printString("FAN board not detected.\r\n");
		printPString("%% FAN control IC not detected.\r\n");
	else
	{
		//printString("FAN board detected.\r\n");
		printPString("%% FAN control IC detected.\r\n");
		uint8_t data, reg;
		// Set Fans to Full off
		reg=FANREG_CONFIG; data=(FANCONFIG_OPENLOOP << 4) | (FANCONFIG_TACH12V << 3) | (FANCONFIG_DIV2 << 0);
		i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1);
		
		// Configure Pins
		reg=FANREG_GPIODEF; //data=FANGPIO4_HIGH | FANGPIO3_HIGH | FANGPIO2_LOW | FANGPIO1_FULLONIN | FANGPIO0_ALERTOUT;
		data=FANGPIO4_HIGH | FANGPIO3_HIGH | FANGPIO2_HIGH | FANGPIO1_FULLONIN | FANGPIO0_ALERTOUT;
		i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1);
		
		// Enable Tacho Alarms
		reg=FANREG_ALARMENABLE; //data=0x00; //data=FANALARM_TACH | FANALARM_MIN | FANALARM_MAX;
		data = FANALARM_GPIO2;
		i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1);
		
		// Set Fan Speed
		reg=FANREG_DAC; data=0; // set fan speed to max value
		i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1);
		
		//Set count time to 0.5 (max range 15 kRPM)
		reg=FANREG_COUNT;
		data = FANCOUNT_0_5;
		i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR,  &reg, 1, &data, 1);
		
		
		//sprintf(x, "%d 1 %X\r\n", i, reg);
		//printString(x);
		//_delay_ms(1000);
	}
}

void setFanSpeedOpenLoop(uint8_t spdRange, uint8_t mode) {
	if (setOperatingMode(FANCONFIG_OPENLOOP) != STATUS_OK)
	{
		//printString("Error setting fan mode\r\n");
		return;
	}
	for (int i = 0; i < FAN_COUNT; ++i)
		fan_target_rpm[i] = spdRange;
	uint8_t data, reg;
	reg=FANREG_DAC; data = FAN_DAC_MAX - spdRange; // 0 is max value (255-255)
	if(i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR, &reg, 1, &data, 1) == TWI_SUCCESS)
	{
		if (mode)// != RESULT_QUIET)
		{
			printf("V %d\r\n", spdRange);
		}
	}
	else
	{
		if (mode) printPString("%% Set FAN speed error!");
	}
}

void setFanSpeedOpenLoopRpm(uint16_t rpm, uint8_t mode)
{
	if (setOperatingMode(FANCONFIG_CLOSEDLOOP) != STATUS_OK)
	{
		//printString("Error setting fan mode\r\n");
		return;
	}
	for (int i = 0; i < FAN_COUNT; ++i)
		fan_target_rpm[i] = rpm;
	uint8_t reg = FANREG_SPEED;
	uint32_t prescaler = 2;
	uint8_t data = ((uint32_t) (30  * prescaler * 1984) / rpm) - 1;	
	if(i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR, &reg, 1, &data, 1) == TWI_SUCCESS)
	{
		if (mode)
		{
			printf("R %d", rpm);
		}	
	}
	else
	{
		if (mode) printPString("%% Set RPM FAN speed error!");
	}
}

int readFansSpeedRpm(uint8_t fan, unsigned *value)
{
	uint8_t tacho_reg[] = {FANREG_TACH0, FANREG_TACH1, FANREG_TACH2};
	uint8_t data;
	uint8_t time_inv = 2; // 1/counter time, e.g. 4 for ct=0.25
	if(i2c_read(&EXTERNAL_I2C, FANCTRL_ADDR, &tacho_reg[fan], 1, &data, 1) == TWI_SUCCESS)
	{
		*value = data*30*time_inv;
		return STATUS_OK;
	}
	else {
		return ERR_IO_ERROR;
	}
}

int setOperatingMode(uint8_t mode) {
	uint8_t addr, data, reg;
	addr = FANCTRL_ADDR;
	reg = FANREG_CONFIG;
	
	if(i2c_read(&EXTERNAL_I2C, addr, &reg, 1, &data, 1) == TWI_SUCCESS)
	{
		data &= ~(0x03 << 4);
		if (mode <= FANCONFIG_OPENLOOP) //if mode is wrong - full on
			data |= (mode << 4); //Open-loop operation
		if(i2c_write(&EXTERNAL_I2C, FANCTRL_ADDR, &reg, 1, &data, 1) == TWI_SUCCESS)
		{
			return STATUS_OK;
		}
		else
		{
			printPString("%% Set Operating Mode error!");
			return ERR_IO_ERROR;
		}
	}
	else {
		printPString("%% FAN board not detected!\r\n");
		return ERR_IO_ERROR;
	}
}

int readOperatingMode(uint8_t *mode) {
	uint8_t addr, data, reg;
	addr = FANCTRL_ADDR;
	reg = FANREG_CONFIG;
	
	if(i2c_read(&EXTERNAL_I2C, addr, &reg, 1, &data, 1) == TWI_SUCCESS)
	{
		*mode = (data >> 4) & 0x3;
		return STATUS_OK;
	}
	else {
		printPString("%% FAN board not detected!\r\n");
		return ERR_IO_ERROR;
	}
}


uint8_t readFanAlarmStatusReg(uint8_t *value) //added by Stanislav Chystiakov
{
	uint8_t addr, reg, data;
	addr = FANCTRL_ADDR;
	reg = FANREG_ALARM;
	
	if(i2c_read(&EXTERNAL_I2C, addr, &reg, 1, &data, 1) == TWI_SUCCESS)
	{
		//*value = data & 0x07;
		*value = data;
		
		//printf("Alarm Status Register = 0x%X\r\n", data); //use this to show value of Alarm Status Register
		return STATUS_OK;
	}
	else {
		//printString("FAN board not detected!\r\n");
		return ERR_IO_ERROR;
	}
}

void printFanInfoHeader(char mode)
{
	if (mode == 'R')
	printPString("fanNr mode alarm current targetValue currentSpeed\r\n");
	else if (mode == 'V')
	printPString("fanNr mode alarm current targetValue currentSpeed\r\n");
	else
	nop();
}

//void printFanInfo(uint8_t fan)
void printFanInfo(void)
{
	int result = 0;
	unsigned current, rpm;
	uint8_t alarm, mode;
	
	uint8_t shuntval=getFanShuntVal();
	if ((shuntval>0) && (shuntval<255)) {
		printf("%% Shunt value is: %d mOhm\r\n", shuntval);
	}
	
	for (uint8_t fan = 0; fan < FAN_COUNT; ++fan)
	{
		result |= readOperatingMode(&mode);
		result |= readFanAlarmStatusReg(&alarm);
		result |= readFanCurrent_mA(fan, &current);
		result |= readFansSpeedRpm(fan, &rpm);
		// Clip Low currents
		if (current <= 3) current = 0;
		if (rpm <= 60) rpm = 0;
		
		if (result == STATUS_OK)
		{
			char mode_c = (mode == FANCONFIG_FULLON) ? '1' : (mode == FANCONFIG_SOFTWAREOFF) ? '0' :
			(mode == FANCONFIG_CLOSEDLOOP) ? 'R' : 'V';
			if (fan == 0) printFanInfoHeader(mode_c);
			
			printf(" %d   | %c  | %1x	| %u	| %u	    | %u\r\n", fan, mode_c, alarm, current, fan_target_rpm[fan], rpm);
		}
		else
		printPString("%% Communication failed\r\n");
	}
}

void control_fan_state(void)
{
	//char message[80];
	static uint8_t channelStatePrev = -1;
	uint8_t channelState = get_channel_state();
	
	//sprintf(message, "chState = %d\r\n", chState); //use this to show value of Alarm Status Register
	//printString(message);
	if(channelState != channelStatePrev)		// check for channel state change
	{
		channelStatePrev = channelState;		// remember channel state
		if(channelState)						// at least one channel is on -> turn fans on
		{
			set_previos_fan_speed();
			forceFanOff = 0;
			//printString("Fans on\r\n");
		}
		else									// all channels off -> turn fans off
		{
			forceFanOff = 1;
			//printString("Fans off\r\n");
			setFanSpeedOpenLoop(FAN_SPEED_V_MODE_ZERO, RESULT_QUIET);
		}
	}
}

uint8_t get_channel_state(void)
{
	uint8_t chState = 0;
	uint8_t i;
	for(i=0; i<8; i++)
	{
		chState |= channels[i].actual_state;
	}
	return chState;
}

void write_fan_parameters(uint8_t mode, uint16_t speedRange)
{
	eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_LOW, speedRange & 0xFF);
	eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_HIGH, (speedRange >> 8) & 0xFF);
	eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_MODE, mode);
}