﻿/**
 * @file mgmt_protocol_master.c
 * @brief Management protocol between TMCB2 and FRED boards support library 
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "i2c_lib.h"
#include "twim.h"
#include "mgmt_protocol_master.h"
#include "usart_lib.h"
#include "telnet_protocol.h"
#include "power_monitor.h"
#include "cpld_protocol.h"
#include "errors.h"
#include "eeprom.h"

#include <stdio.h>
#include <util/delay.h>
#include <avr/eeprom.h>


FpgaConfigureState tmcb_fpga_state = Not_configured;
uint16_t tmcb_status = 0;

bool tmcb_fpga_poweroff_request = false;
bool tmcbJumperState = false;

/**
 * \brief Read register on mgmt bus
 * 
 * \param reg Register
 * \param data Pointer to data
 * 
 * \return int Status
 */
int mgmtReadRegister( uint8_t reg, uint16_t *data)
{
	int8_t result = 0;
	if ( (result = i2c_read(&EXTERNAL_I2C, MANAGE_ADDR,  &reg, 1, (uint8_t*) data, 2)) == TWI_SUCCESS)
	{
		setCpldErrorLED(false);
		increment_tmcb_mgmt(result);
		return STATUS_OK;
	}
	else
	{
		setCpldErrorLED(true);
		increment_tmcb_mgmt(result);
		return ERR_IO_ERROR;
	}
}

int mgmtWriteRegister( uint8_t reg, uint16_t data )
{
	uint8_t replies = REPLIES_TO_FAIL;
	twi_package_t packet_write;
	packet_write.addr_length  = 1;					// TWI slave memory address data size
	packet_write.addr[0]	  = reg;
	packet_write.chip         = MANAGE_ADDR;		// TWI slave bus address
	packet_write.buffer       = (void *)&data;		// transfer data source buffer
	packet_write.length       = 2;					// transfer data size (bytes)
	
	int8_t result = 0;
	while(replies--)
	{
		if((result = twi_master_write(&EXTERNAL_I2C, &packet_write)) == TWI_SUCCESS)
		{
			setCpldErrorLED(false);
			increment_tmcb_mgmt(result);
			return STATUS_OK;
			
		}
	}
	setCpldErrorLED(true);	
	return ERR_TIMEOUT;
}

/**
 * \brief Read voltage from TMCB2
 * 
 * \param sensor
 * 
 * \return uint16_t
 */
uint16_t mgmtReadVoltage( uint8_t sensor )
{
	uint16_t result = 0;
	mgmtReadRegister(REG_VOLTAGE_FIRST | (sensor & 0x0F), &result);
	return result;
}


/**
 * \brief Handling hot-plug management of FRED board
 * 
 * \return void
 */
void processTmcbMgmt(void)
{
	#ifdef DBG_LEVEL_VERBOSE
		printPString("% Reading TMCB status\r\n");
	#endif
	tmcb_status = 0;
	getTmcbStatus(&tmcb_status);
}

void tmcbEnableFpga(void)
{
	mgmtWriteRegister(REG_FPGA_CONTROL, REG_FPGA_CONTROL_ENABLE);
	setTmcbBlueLed(REG_BLUE_LED_OFF);
	tmcb_fpga_state = Configured;
}

void tmcbRequestDisableFpga(void)
{
	
	#ifdef DBG_LEVEL_VERBOSE
		printPString("% Requested tmcb disable\r\n");
	#endif
		
	if ((tmcb_status & REG_STATUS_ALLOW_DISABLE) != 0)
	{
		mgmtWriteRegister(REG_FPGA_CONTROL, REG_FPGA_CONTROL_DISABLE);
		tmcb_fpga_state = Disabling;
		setTmcbBlueLed(REG_BLUE_LED_BLINK);
		tmcb_fpga_poweroff_request = true;
	}
}

int getTmcbStatus( uint16_t *status )
{
	int res = mgmtReadRegister(REG_TMCB_STATUS, status);
	return res;
}

int setTmcbBlueLed( uint16_t value )
{
	return mgmtWriteRegister(REG_BLUE_LED, value);
}

/**
 * \brief Indicates if channel can be powered on
 * In operation without TMCB (jumper off) - always allow
 * 
 * \return bool
 */
bool canPowerUp( void )
{
	if (tmcbJumperState)
	{
	//	if (!(tmcb_status & REG_STATUS_ALLOW_ENABLE))
	//		printf("%%%% TMCB does not allow power up\r\n");
		return (tmcb_status & REG_STATUS_ALLOW_ENABLE);
	}
	return true;
}

/**
 * \brief Indicates if channel can be powered down
 * In operation without TMCB (jumper off) - always allow
 * 
 * \return bool
 */
bool canPowerDown( void )
{
	if (tmcbJumperState)
	{
		//if (!(tmcb_status & REG_STATUS_ALLOW_DISABLE))
		//	printf("%%%% TMCB does not allow power down\r\n");
		return (tmcb_status & REG_STATUS_ALLOW_DISABLE);
	}
	return true;
}

/**
 * \brief Test management protocol and print results on serial console
 * 
 * 
 * \return void
 */
void testMgmtProtocol(void)
{
	char msg[80];
	const char *names[] = {
		"REG_TEST", "REG_POWER_GOOD", "REG_FPGA_CONTROL",
		"REG_FPGA_PROM", "REG_TEMPERATURE", "REG_FMC", "REG_SW",
		"REG_TEMPERATURE_STATUS"
	};
	const char* adc_channel_name[] = { "+15VD",
		"+12VA",
		"+10VA",
		"+5VA",
		"+5VD",
		"+3V3D",
		"+2V5A",
		"-5VA",
		"-10VA",
		"-12VA",
		"-15VD"
	};
	uint16_t data;
	printPString("% *****************\r\n ");
	for (int i = 0; i < sizeof(names)/sizeof(names[0]); ++i)

	{
		mgmtReadRegister(i, &data);
		sprintf_P(msg, PROGMEM_STRING("Register %s value %X\r\n"), names[i], data);
		printString(msg);
	}
	for (int i = 0; i < sizeof(adc_channel_name)/sizeof(adc_channel_name[0]); ++i)

	{
		sprintf_P(msg, PROGMEM_STRING("Voltage %s value %X\r\n"), adc_channel_name[i], mgmtReadVoltage(i));
		printString(msg);		
	}
}

bool tmcbIsConfigured( void )
{
	return tmcb_fpga_state == Configured;
}

bool tmcbJumperOn()
{
	return ioport_pin_is_high(CPLD_TMCB_JUMPER);
}

bool tmcbHandleOn( void )
{
	return tmcb_status & REG_STATUS_AMC_HANDLE;
}

void checkTmcbJumperPresence(void)
{
	char savedJumperMode;
	savedJumperMode = eeprom_read_byte(EEPROM_ADDR_JUMPER_PRESENCE);
	printPString("%% TMCB mode: ");
	switch (savedJumperMode)
	{
		case TMCB_PRESENCE_PRESENT: tmcbJumperState = TMCB_JUMPER_PRESENT; printPString("%% force present\r\n"); break;
		case TMCB_PRESENCE_ABSENT: tmcbJumperState = TMCB_JUMPER_ABSENT; printPString("%% force absent\r\n"); break;
		default: // == TMCB_PRESENCE_HWJUMPER; also goes here when value is not set
		{ 
			tmcbJumperState = tmcbJumperOn();
			if(tmcbJumperState) printPString("%% HW jumper (present)\r\n");
			else printPString("%% HW jumper (absent)\r\n");	
		}
	}
}

void writeTmcbJumperMode(char mode)
{
	eeprom_write_byte(EEPROM_ADDR_JUMPER_PRESENCE, mode);
}