﻿/**
 * @file power_monitor.c
 * @brief Library for communication with power monitors and handling channel info
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Piotr Perek (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "power_monitor.h"
#include "i2c_lib.h"
#include "cpld_protocol.h"
#include "mgmt_protocol_master.h"
#include "eeprom.h"
#include "hwrevision.h"
#include "errors.h"
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>

RelayChannel channels[8];


uint8_t reg_control_addr = REG_CONTROL;
uint8_t reg_control_data_continuous = 0x01;
uint8_t reg_control_data_snapshot = 0x81;

uint8_t unconditional_mask = 0;
uint8_t hotplug_mask = 0;

bool power_on_request = false,
	power_down_request = false,
	power_off_request = false;
/**
 * \brief Initialize all channels
 * 
 * \param 
 * 
 * \return void
 */
void initChannelStructures( void )
{
	//P15VD
	channels[0].i2c_addr = I2C_PWR_MON_P15VD_ADDR;
	channels[0].negative = false;
	channels[0].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL0;
	//P15VA
	channels[1].i2c_addr = I2C_PWR_MON_P15VA_ADDR;
	channels[1].negative = false;
	channels[1].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL1;
	//P15VH
	channels[2].i2c_addr = I2C_PWR_MON_P15VH_ADDR;
	channels[2].negative = false;
	channels[2].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL2;
	//P7VA
	channels[3].i2c_addr = I2C_PWR_MON_P7VA_ADDR;
	channels[3].negative = false;
	channels[3].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL3;
	//N15VD
	channels[4].i2c_addr = I2C_PWR_MON_N15VD_ADDR;
	channels[4].negative = true;
	channels[4].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL4;
	//N15VA
	channels[5].i2c_addr = I2C_PWR_MON_N15VA_ADDR;
	channels[5].negative = true;
	channels[5].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL5;
	//N7VA
	channels[6].i2c_addr = I2C_PWR_MON_N7VA_ADDR;
	channels[6].negative = true;
	channels[6].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL6;
	//N15VH
	channels[7].i2c_addr = I2C_PWR_MON_N15VH_ADDR;
	channels[7].negative = true;
	channels[7].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL7;
	
	readRelays();

	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		channels[channel].alarm_mask = 
			(eeprom_read_byte(EEPROM_ADDR_LIMIT_MASK) & (1 << channel)) ? true : false;
		channels[channel].overcurrent_alarm_status = 0;
		channels[channel].overcurrent_alarm_counter = 0;
		channels[channel].alarm_pin0_status = false;
		channels[channel].alarm_pin1_status = false;
		channels[channel].current_limit = eeprom_read_dword(channels[channel].eeprom_addr);
		channels[channel].on_off_state = channels[channel].actual_state;
		channels[channel].i2c_error = false;
		channels[channel].state_changed = true;
		eeprom_chname_read(channel, channels[channel].name);
	}
	
	//Fred 3m - only positive channels
	if ((get_hwrevision() == FRED3MC) || (get_hwrevision() == FRED3MD))
	{
		for (int i = 0; i < CHANNEL_COUNT; ++i)
			channels[i].negative = false;
	}

	
	//enable portC interrupts
	PORTC.INT0MASK = ioport_pin_to_mask(SWITCH_UP) | ioport_pin_to_mask(SWITCH_DOWN);
	PORTC.INT1MASK = ioport_pin_to_mask(CPLD_INTERRUPT);
	PORTC.INTCTRL = (PORT_INT0LVL_HI_gc | PORT_INT1LVL_HI_gc);
}


///ISR for switch
ISR(PORTC_INT0_vect)
{
	if (ioport_pin_is_high(SWITCH_UP))
	{
		power_on_request = true;
	}
	else if (ioport_pin_is_high(SWITCH_DOWN))
	{
		power_down_request = true;
	}
}

///ISR for CPLD interrupt
ISR(PORTC_INT1_vect)
{
	if (ioport_pin_is_high(CPLD_INTERRUPT))
	{
		readRelays();
	}
}


/**
 * \brief Update channel
 * 
 * \param channel Channel
 * \param nr Channel number
 * \param timerTriggerred Flag set if the function is invoked by timer event, not by commands
 * 
 * \return void
 */
void processChannel( RelayChannel *channel, int nr, bool timerTrigerred )
{
	int8_t status = 0;
	if((status = i2c_write(&LOCAL_I2C, channel->i2c_addr, &reg_control_addr, 1, &reg_control_data_continuous, 1)) == TWI_SUCCESS)
	{
		increment_i2c_power_monitor(nr, status);
		status = i2c_power_monitor_read_voltage(&LOCAL_I2C, channel->i2c_addr, &(channel->voltage));
		increment_i2c_power_monitor(nr, status);
		status = i2c_power_monitor_read_current(&LOCAL_I2C, channel->i2c_addr, &(channel->current));
		increment_i2c_power_monitor(nr, status);
		status = i2c_power_monitor_read_power(&LOCAL_I2C, channel->i2c_addr, &(channel->power));
		increment_i2c_power_monitor(nr, status);
		channel->i2c_error = false;
	}
	else
	{
		increment_i2c_power_monitor(nr, status);
		channel->voltage = 0;
		channel->current = 0;
		channel->i2c_error = true;
	}
	if (channel->overcurrent_alarm_status & channel->actual_state)
	{
		channel->overcurrent_alarm_status = 0;
	}
	if ((channel->current >= channel->current_limit) && channel->alarm_mask)
	{
		if (timerTrigerred)
		{
			channel->overcurrent_alarm_counter++;
		}
		if (channel->overcurrent_alarm_counter > OVERCURRENT_LIMIT_COUNTER)
		{	
			channel->state_changed = true;
			channel->overcurrent_alarm_status = true;
			#ifdef DBG_LEVEL_ERROR
			printf("%%%%%% Alarm triggered on channel %d\r\n", nr);
			#endif
		}
	}
	else
	{
		channel->overcurrent_alarm_counter = 0;
	}
	if ((readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) & CPLD_ADDR_ALARMPIN1_TRIGGERED) 
		&& ((readRegister(CPLD_ADDR_ALARMPIN0_MASK)) & (1 << nr)))
	{
		channel->alarm_pin0_status = true;
		channel->on_off_state = 0;
		channel->state_changed = true;
	}
	else
	{
		channel->alarm_pin0_status = false;
	}
	if ((readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) & CPLD_ADDR_ALARMPIN2_TRIGGERED)
		&& ((readRegister(CPLD_ADDR_ALARMPIN1_MASK)) & (1 << nr)))
	{
		channel->alarm_pin1_status = true;
		channel->on_off_state = 0;
		channel->state_changed = true;
	}
	else
	{
		channel->alarm_pin1_status = false;
	}
	
}

/**
 * \brief Update all channels
 * \param timerTriggerred Flag set if the function is invoked by timer event, not by commands
 * 
 * \return void
 */
void processChannels(bool timerTriggerred)
{
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		processChannel(&channels[channel], channel, timerTriggerred);
	}
	readRelays();
}

uint8_t getAlarmMask()
{
	uint8_t res = 0;
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		res |= ((channels[channel].alarm_mask ? 1 : 0) << channel);
	}
	return res;
}

void setAlarmMask(uint8_t alarmMask)
{
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		if ((alarmMask >> channel) & 1)
		channels[channel].alarm_mask = true;
		else
		channels[channel].alarm_mask = false;
	}
	eeprom_write_byte(EEPROM_ADDR_LIMIT_MASK, alarmMask);
}

/**
 * \brief Set current limit for a channel, not greater than MAX_LIMIT
 * 
 * \param channel Channel number
 * \param limit Limit
 * 
 * \return void
 */
void setCurrentLimit(int channel, uint32_t limit)
{
	if (limit < MAX_CURRENT_LIMIT)
	channels[channel].current_limit = limit;
	else
	channels[channel].current_limit = MAX_CURRENT_LIMIT;
	eeprom_update_cur_limit(channel);
}

/**
 * \brief Read state of relays
 * 
 * \param 
 * 
 * \return void
 */
void readRelays( void )
{
	int channel_mask = getRelaysState();
	for (int i = 0; i < CHANNEL_COUNT; ++i)
	{
		channels[i].actual_state = (channel_mask & (1 << i))? true : false;
	}
}

/**
 * \brief Set requested state on relays
 * 
 */
void updateRelays( void )
{
	int channel_mask = readRegister(CPLD_ADDR_RELAY_STATUS);
	bool changed = false;
	for (int power_channel = 0; power_channel < CHANNEL_COUNT; ++power_channel)
	{
		if (channels[power_channel].state_changed )
		{
			if (channels[power_channel].on_off_state & !(channels[power_channel].overcurrent_alarm_status))
			{
				//Check if TMCB allows to switch power on
				//If TMCB jumper not present, switch on
				//Always switch channels from unconditional mask regardless of tmcb
				if (canPowerUp() || unconditional_mask & (1 << power_channel))
				{
					channel_mask |= (1 << power_channel);
					changed = true;
				}
			}
			else
			{
				//Always switch channels from unconditional mask regardless of tmcb
				if (canPowerDown() || unconditional_mask & (1 << power_channel))
				{
					channel_mask &= ~(1 << power_channel);
					changed = true;
				}
			}
			if (channels[power_channel].overcurrent_alarm_status)
			{
				channel_mask &= ~(1 << power_channel);
				changed = true;
			}
		}
		channels[power_channel].state_changed = false;
	}
	if (changed)
	{
		setRelaysState(channel_mask);
	}
}

uint8_t getPowerState()
{
	return readRegister(CPLD_ADDR_RELAY_CONTROL);
}

void clearOvercurrentAlarm( int channel )
{
/*	if (channels[channel].state & channels[channel].overcurrent_alarm_status)
	{
		channels[channel].state = false;
	}*/
	channels[channel].overcurrent_alarm_status = false;
}


/**
 * \brief Force read of channels
 * 
 * 
 * \return void
 */
void channelsSnapshot()
{
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		if(!i2c_power_monitor_start_snapshot_measurement(&channels[channel]))
		{
			channels[channel].i2c_error = true;
		}
	}
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		int8_t status = 0;
		if (channels[channel].i2c_error) continue;
		i2c_power_monitor_wait_for_snapshot(&channels[channel]);
		increment_i2c_power_monitor(channel, status);
		status = i2c_power_monitor_read_voltage(&LOCAL_I2C, channels[channel].i2c_addr, &(channels[channel].voltage));
		increment_i2c_power_monitor(channel, status);
		status = i2c_power_monitor_read_current(&LOCAL_I2C, channels[channel].i2c_addr, &(channels[channel].current));
		increment_i2c_power_monitor(channel, status);
		status = i2c_power_monitor_read_power(&LOCAL_I2C, channels[channel].i2c_addr, &(channels[channel].power));
		increment_i2c_power_monitor(channel, status);
	}
}

/**
 * \brief Power on - with switch or command
 * 
 * \param 
 * 
 * \return void
 */
void powerBoxOn( void )
{
	#ifdef DBG_LEVEL_VERBOSE
	 printPString("% Box power on - switch\r\n");
	#endif
		
	power_on_request = false;
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		if (unconditional_mask & (1 << channel) || hotplug_mask & (1 << channel))
		{
			channels[channel].on_off_state = true;
			channels[channel].state_changed = true;
		}
	}
	readRelays();
	if (tmcbJumperState)
		powerOnTmcbHandle();
	updateRelays();
}

void powerOff( void )
{
	#ifdef DBG_LEVEL_VERBOSE
	 printPString("% Box power off - TMCB disabled\r\n");
	#endif
	
	power_off_request = false;
	
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		if (unconditional_mask & (1 << channel))
		{
			channels[channel].on_off_state = false;
			channels[channel].state_changed = true;
		}
	}
		
	if (tmcbJumperState)
	{
		if (!canPowerDown())
		{
			updateRelays();
			return;
		}
	}
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
			channels[channel].on_off_state = 0;
			channels[channel].state_changed = true;
	}	
	if (tmcbJumperState)
		powerOffTmcbHandle();
}

void powerBoxDown(void)
{
	#ifdef DBG_LEVEL_VERBOSE
	  printPString("% Box power down - switch\r\n");
	#endif
		
	power_down_request = false;
	if (tmcbJumperState)
	{
		tmcbRequestDisableFpga();
		power_off_request = true;
	}
	else
	{
		power_off_request = true;
	}
}

void setUnconditionalMask( uint8_t mask )
{
	unconditional_mask = mask;
	eeprom_write_unconditional_mask(mask);
}

void setHotplugMask( uint8_t mask )
{
	hotplug_mask = mask;
	eeprom_write_hotplug_mask(mask);
}

void loadPowerMasks(void)
{
	unconditional_mask = eeprom_read_unconditional_mask();
	hotplug_mask = eeprom_read_hotplug_mask();
}

/**
 * \brief Tells if box is powered
 * 
 * box powered - at least one power channel is on
 * \param 
 * 
 * \return bool
 */
bool boxPowered(void)
{
	bool powered = false;
	for (int i = 0; i < CHANNEL_COUNT; ++i)
		if (channels[i].actual_state == true)
			powered = true;
	return powered;
}

/**
 * \brief Power on the box when TMCB handle is on 
 * Required: box on
 * Only channels specified in HOTPLUG MASK 
 * Start TMCB fpga
 * 
 */
void powerOnTmcbHandle( void )
{
	#ifdef DBG_LEVEL_VERBOSE
		printPString("% Enabling TMCB\r\n");
	#endif
		
	setTmcbBlueLed(REG_BLUE_LED_OFF);
	//Required: box powered
	if (!boxPowered())
		return;
	if (canPowerUp())
	{
		tmcb_fpga_poweroff_request = false;
		//Switch on according to HOTPLUG MASK
		for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
		{
			if (hotplug_mask & (1 << channel))
			{
				channels[channel].on_off_state = true;
				channels[channel].state_changed = true;
			}
		}
		updateRelays();
		tmcbEnableFpga();
	}

}



/**
 * \brief 
 * 
 * \param 
 * 
 * \return void
 */
void powerOffTmcbHandle( void )
{
	if (canPowerDown())
	{
		tmcb_fpga_poweroff_request = false;
		//Switch off according to HOTPLUG MASK
		for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
		{
			if (hotplug_mask & (1 << channel))
			{
				channels[channel].on_off_state = 0;
				channels[channel].state_changed = true;
			}
		}	
		setTmcbBlueLed(REG_BLUE_LED_ON);
		tmcb_fpga_state = Not_configured;
	}
}
