﻿/**
 * @file telnet_protocol.c
 * @brief Telnet protocol for Xport
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "telnet_protocol.h"
#include "power_monitor.h"
#include "usart_lib.h"
#include "cpld_protocol.h"
#include "user_board.h"
#include "i2c_lib.h"
#include "mgmt_protocol_master.h"
#include "fan_lib.h" // added by Stanislav Chystiakov
#include "fan_current_monitor_lib.h" // added by Stanislav Chystiakov
#include "eeprom.h"
#include "errors.h"
#include "hwrevision.h"
#include "version.h"

#include <util/delay.h>
#include <usart_serial.h>
#include <avr/eeprom.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>

volatile bool bufferReady = false;
volatile int inputBufferCounter = 0;
volatile char inputBuffer[INPUT_BUFFER_SIZE] = {0};
char uppercase(char c);


void clearBuffer( void )
{
	inputBuffer[0] = '\0';
	inputBufferCounter = 0;
	bufferReady = false;
	//enable UART interrupt
	usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_LO);
	//(USART_SERIAL)->CTRLA |= (0x1 << 4);
}

void telnetPutChar(int ch)
{
	if (ch == '\b' && inputBufferCounter > 0)
	{
		inputBufferCounter--;
	}
	if (ch == '\r')
	{
		inputBuffer[inputBufferCounter] = '\0';
		bufferReady = true;
		//disable UART interrupt
		usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_OFF);
	}
	if (!isprint(ch))
	{
		return;
	}
	if (inputBufferCounter < INPUT_BUFFER_SIZE)
	inputBuffer[inputBufferCounter++] = tolower(ch);
	else
	{
		inputBuffer[INPUT_BUFFER_SIZE-1]='\0';
		//disable UART interrupt
		usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_OFF);
		bufferReady = true;
	}
}

void channelsOn()
{
	power_on_request = true;
	processChannels(false);
}

void channelsOff()
{
	power_down_request = true;
	processChannels(false);
}

void printAlarmStatus(void)
{
	printf("%02X",readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED));
}

void printTmcbPinVoltages(void)
{
	uint16_t data = 0;
	if (mgmtReadRegister(REG_TMCB_PIN_VOLTAGE, &data) != ERR_IO_ERROR)
	{
		for (int pin = 1; pin < 20; pin += 2)
		{
			printf("Pins %d-%d - %sV\r\n", pin, pin+1,
			(data & (1 << ((pin-1) / 2))) ? "3.3" : "5");
		}
	}
	else
	printPString("%%%% Error in communication with TMCB!\r\n");
}

/**
 * \brief 
 * Parses the command from inputBuffer
 * \param 
 * 
 * \return void
 */
void parseCommand( void )
{
	///Supported commands:
	if (strstr( (char*) inputBuffer, "readchan") == inputBuffer)
	{
		///READ CHANNEL
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "readchan %u", &channel) != 1 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			channelsSnapshot();
			printChannelHeader();
			printChannel(channel);
		}
	}
	else if (strstr( (char*) inputBuffer, "readall") == inputBuffer)
	{
		///READ ALL CHANNELS
		channelsSnapshot();
		printChannelHeader();
		for (int i = 0; i < CHANNEL_COUNT; ++i)
		{
			printChannel(i);
		}
	}
	else if (strstr( (char*) inputBuffer, "setpwrstates") == inputBuffer)
	{
		///SET POWER STATES
		unsigned powerStateWord, channelMask;
		if (sscanf( (char*) inputBuffer, "setpwrstates %2X %2X", &powerStateWord, &channelMask) != 2)
		{
			printBadArgument();
		}
		else
		{
			for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
			{
				if ( (channelMask >> channel) & 0x01 )
				{
					if ( (powerStateWord >> channel) & 0x01)
					{
						channels[channel].on_off_state = true;
						channels[channel].state_changed = true;
					}
					else
					{
						channels[channel].on_off_state = false;
						channels[channel].state_changed = true;							
					}
				}
			}
			processChannels(false);
			updateRelays();
			printPowerStateWord();
		}
	}
	else if (strstr( (char*) inputBuffer, "chanon") == inputBuffer)
	{
		///TURN CHANNEL ON
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "chanon %u", &channel) != 1 || channel > CHANNEL_COUNT - 1)
		{
			printBadArgument();
		}
		else
		{
			channels[channel].on_off_state = true;
			channels[channel].state_changed = true;
			updateRelays();
			printPowerStateWord();
		}
	}
	else if (strstr( (char*) inputBuffer, "chanoff") == inputBuffer)
	{
		///TURN CHANNEL OFF
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "chanoff %u", &channel) != 1 || channel > CHANNEL_COUNT - 1)
		{
			printBadArgument();
		}
		else
		{
			channels[channel].on_off_state = 0;
			channels[channel].state_changed = true;
			updateRelays();
			printPowerStateWord();
		}
	}
	else if (strstr( (char*) inputBuffer, "setchanname") == inputBuffer)
	{
		unsigned int channel;
		char name[21];
		if (sscanf( (char*) inputBuffer, "setchanname %u %20s", &channel, name) != 2 || channel > CHANNEL_COUNT - 1)
		{
			printBadArgument();
		}
		else
		{ 
			strncpy(channels[channel].name, name, EEPROM_CHNAME_LEN);
			channels[channel].name[EEPROM_CHNAME_LEN] = '\0'; 
			eeprom_chname_write(channel, name);
		}
	}
	else if (strstr( (char*) inputBuffer, "readcurlimits") == inputBuffer)
	{
		///READ ALL CHANNELS' CURRENT LIMITS
		printLimits();
	}
	else if (strstr( (char*) inputBuffer, "progcurlimit") == inputBuffer)
	{
		///SET CURRENT LIMIT
		uint32_t channel, limit;
		if (sscanf( (char*) inputBuffer, "progcurlimit %"PRIu32" %"PRIu32"", &channel, &limit) != 2 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			setCurrentLimit(channel,limit);
		//	clearOvercurrentAlarm(channel);
			printf("%"PRIu32"", channels[channel].current_limit);
		}
	}
	else if (strstr( (char*) inputBuffer, "poweron") == inputBuffer)
	{
		///TURN ALL CHANNELS ON
		channelsOn();	
		printPowerStateWord();
	}
	else if (strstr( (char*) inputBuffer, "poweroff") == inputBuffer)
	{
		///TURN ALL CHANNELS OFF
		channelsOff();
		printPowerStateWord();
	}
	else if (strstr( (char*) inputBuffer, "setalarmmasks") == inputBuffer)
	{
		///SET ALARM MASKS
		uint8_t alarmMasks, pin0Mask, pin1Mask;
		if (sscanf( (char*) inputBuffer, "setalarmmasks %2"PRIX8" %2"PRIX8" %2"PRIX8,
			&alarmMasks, & pin0Mask, &pin1Mask) != 3)
		{
			printBadArgument();
		}
		else
		{
			setAlarmMask(alarmMasks);
			printAlarmMasks();
			setAlarmPin0Mask(pin0Mask);
			setAlarmPin1Mask(pin1Mask);
			printPinAlarmMasks();
		}
	}
	else if (strstr( (char*) inputBuffer, "readalarms") == inputBuffer)
	{
		///READ ALARMS
		printAlarmStatus();
		
	}
	else if (strstr( (char*) inputBuffer, "clearalarms") == inputBuffer)
	{
		///CLEAR ALARMS
		clearAlarms();
		printAlarmStatus();
	}
	else if (strstr( (char*) inputBuffer, "readalarmmasks") == inputBuffer)
	{
		printAlarmMasks();
		printPinAlarmMasks();
	}
	else if (strstr( (char*) inputBuffer, "setpwrmasks") == inputBuffer)
	{
		///SET POWER MASKS
		unsigned unconditional, hotplug;
		if (sscanf( (char*) inputBuffer, "setpwrmasks %2X %2X", &unconditional, &hotplug) != 2)
		{
			printBadArgument();
		}
		else
		{
			setUnconditionalMask(unconditional);
			setHotplugMask(hotplug);
			printf("Power masks: %02X %02X\r\n", unconditional_mask, hotplug_mask);
		}
	}
	else if (strstr( (char*) inputBuffer, "readpwrmasks") == inputBuffer)
	{
		printf("Power masks: %2X %2X\r\n", unconditional_mask, hotplug_mask);
	}
	else if (strstr( (char*) inputBuffer, "tmcbselectprom") == inputBuffer)
	{
		///SET FPGA PROM
		uint8_t value;
		if (sscanf( (char*) inputBuffer, "tmcbselectprom %2X", &value) != 1)
		{
			printBadArgument();
		}
		else
		{
			uint16_t data = 0;
			if (value)
				data = 1;
			else
				data = 0;
			if (mgmtWriteRegister(REG_FPGA_CONTROL, data) == TWI_SUCCESS)
			{
				printPString("Success\r\n");
			}
			else
				printPString("%%%% Error in communication with TMCB!\r\n");
		}
	}
	else if (strstr( (char*) inputBuffer, "tmcbreadfpgastatus") == inputBuffer)
	{
		///READ FPGA STATUS
		uint16_t data = 0;
		if (mgmtReadRegister(REG_TMCB_STATUS, &data) != ERR_IO_ERROR)
		{
			printf("%02X",(data & REG_STATUS_FPGA_DONE) ? 1 : 0);
		}
		else
		printPString("%%%% Error in reading from TMCB!\r\n");
	}
	else if (strstr( (char*) inputBuffer, "tmcbreloadfpga") == inputBuffer)
	{
		///SET FPGA PROG B
		if (mgmtWriteRegister(REG_FPGA_CONTROL, REG_FPGA_CONTROL_RELOAD) != STATUS_OK)
			printPString("%%%% Error in reading from TMCB!\r\n");
	}
	else if (strstr( (char*) inputBuffer, "tmcbreadpinvoltages") == inputBuffer)
	{
		printTmcbPinVoltages();

	}
	else if (strstr( (char *) inputBuffer, "tmcbhardreset") == inputBuffer)
	{
		powerOffTmcbHandle();
		disableTmcbMP();
		_delay_ms(1000);
		enableTmcbMP();
		powerOnTmcbHandle();
	}
	else if (strstr( (char*) inputBuffer, "settmcbpresence") == inputBuffer) //
	{
		char mode;
		if ((sscanf( (char*) inputBuffer, "settmcbpresence %c", &mode) != 1)
		|| (mode != TMCB_PRESENCE_HWJUMPER && mode != TMCB_PRESENCE_PRESENT && mode != TMCB_PRESENCE_ABSENT))
		{
			printPString("Supported jumper modes - j (HW jumper), p (present), a (absent)\r\n");
			printBadArgument();			
		}		
		else
		{
			writeTmcbJumperMode(mode);
		}

	}
	else if (strstr( (char*) inputBuffer, "exit") == inputBuffer)
	{
		printPString("Command not implemented");
	}
	else if (strstr( (char*) inputBuffer, "version") == inputBuffer)
	{
		printVersion();
	}
	else if (strstr( (char*) inputBuffer, "fwupgrade") == inputBuffer)
	{
		eeprom_write_upgrade_requested();
		//reset CPU
		ccp_write_io((void *)&RST.CTRL, RST_SWRST_bm);
	}
	else if (strstr( (char*) inputBuffer, "progdevicename ") == inputBuffer)
	{
		eeprom_write_box_name((char *) &(inputBuffer[15]));
	}
	else if (strstr( (char*) inputBuffer, "proginvnr ") == inputBuffer)
	{
		eeprom_write_inv_no((char *) &(inputBuffer[10]));
	}
	else if (strstr( (char*) inputBuffer, "printstats") == inputBuffer)
	{
		///Print errors on i2c buses
		print_transmission_statistics();
	}
	else if (strstr( (char*) inputBuffer, "reseti2c"))
	{
		printf("Resetting i2c bus\n");
		i2c_reset_bus(&EXTERNAL_I2C);
		i2c_reset_bus(&LOCAL_I2C);
	}
    else if (strstr( (char*) inputBuffer, "reset") == inputBuffer)
    {
        //reset CPU
        ccp_write_io((void *)&RST.CTRL, RST_SWRST_bm);
    }
	else if (strstr( (char*) inputBuffer, "a0") == inputBuffer)
	{
		writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 
			readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) | 0x1);
		printAlarmStatus();
	}
	else if (strstr( (char*) inputBuffer, "a1") == inputBuffer)
	{
		writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 
			readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) | 0x2);
		printAlarmStatus();
	}
	else if (strstr( (char*) inputBuffer, "help") == inputBuffer)
	{
		///HELP
		printHelp();
	}
	/*--------------------------------------------------------------------------------------------- Commands for FRED Fan  */
		
	else if (strstr( (char*) inputBuffer, "caladc") == inputBuffer) // calibrate ADCs (only for FAN current sensors)
	{
		adcCalibrate();
	}
	else if (strstr( (char*) inputBuffer, "ralarm") == inputBuffer) // read alarm status register
	{
		uint8_t Val =  11;
		readFanAlarmStatusReg(&Val);
		printf("Alarm Status Register = 0x%X\r\n", Val);
	}
	else if (strstr( (char*) inputBuffer, "rmem") == inputBuffer) //
	{
		unsigned int Val;
		if ((sscanf( (char*) inputBuffer, "rmem %u", &Val) != 1))
		{
			printBadArgument();
		}
		else
		{
			printf("EEPROM Register [%d] = %d\r\n", Val, eeprom_read_byte((uint8_t *) Val));
		}		
	}	
	else if (strstr( (char*) inputBuffer, "fbsetshunt") == inputBuffer) // set value of the fan shunt
	{		
		unsigned int Val;
		if ((sscanf( (char*) inputBuffer, "fbsetshunt %u", &Val) != 1))
		{
			printBadArgument();
		}
		else
		{
			if ((Val > 0x00) && (Val < 0xFF))
			{
				setFanShuntVal(Val);
			}
			else printPString("%% Supported shunt value is 1 ... 244 mOhm!\r\n");
		}
	}	
	else if (strstr( (char*) inputBuffer, "fbprogfan") == inputBuffer) /// set FAN speed. Max speed range is 255!
	{	
		unsigned int speedRange;
		char mode;
		if ((sscanf( (char*) inputBuffer, "fbprogfan %c %u", &mode, &speedRange) != 2) || (mode != 'v' && mode != 'r'))
		{
			printPString("Supported modes - R, V\r\n");
			printBadArgument();
		}
		else if (mode == 'v' && ((speedRange >= 0) && speedRange <= 255))
		{
			setFanSpeedOpenLoop(speedRange, RESULT_CLI);
			write_fan_parameters(FAN_SPEED_MODE_VOLTAGE, speedRange);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_LOW, speedRange & 0xFF);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_HIGH, (speedRange >> 8) & 0xFF);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_MODE, FAN_SPEED_MODE_VOLTAGE);
		}
		else if (mode == 'r')
		{
			setFanSpeedOpenLoopRpm(speedRange, RESULT_CLI);
			write_fan_parameters(FAN_SPEED_MODE_RPM, speedRange);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_LOW, speedRange & 0xFF);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_HIGH, (speedRange >> 8) & 0xFF);
			//eeprom_write_byte(EEPROM_ADDR_FAN_SPEED_MODE, FAN_SPEED_MODE_RPM);
		}
		else
		{
			printPString("For Voltage mode - value in range 0-255");
			printBadArgument();
		}
	}
	else if (strstr( (char*) inputBuffer, "fbsetaverage") == inputBuffer) 
	{
		/// set average
		unsigned int avrgVal;
		if ((sscanf( (char*) inputBuffer, "fbsetaverage %u", &avrgVal) != 1) || (avrgVal < 1) || (avrgVal > 255))
		{
			printBadArgument();
		}
		else
		{
			setAverage(avrgVal);
		}
	}	
	else if ( (strstr( (char*) inputBuffer, "fbreadfansinfo") == inputBuffer) ||
				(strstr( (char*) inputBuffer, "rfi") == inputBuffer)) // read FAN measurements
	{
		///read info for all fans
		//printFanInfoHeader();
		//for (int i = 0; i < FAN_COUNT; ++i)
	    //printFanInfo(i);
		printFanInfo();
	}
	else if (strstr( (char*) inputBuffer, "fbsetioportdir" ) == inputBuffer) 
	{
		/// set Fanboard IO port direction
		unsigned int portDir;
		char port;
		if ((sscanf( (char*) inputBuffer, "fbsetioportdir %c %x", &port, &portDir) != 2) || (portDir > 255)) // input in HEX
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			setPortDirection(FB_PORTB_, portDir);
		}
		else if (port == 'd')
		{
			setPortDirection(FB_PORTD_, portDir);
		}
		else{
			printPString("Supported ports - B and D");
			printBadArgument();
		}
	}
	else if (strstr( (char*) inputBuffer, "fbsetioportlevel") == inputBuffer) 
	{
		///set Fanboard IO port level
		unsigned int portLev;
		char port;
		if ((sscanf( (char*) inputBuffer, "fbsetioportlevel %c %x", &port, &portLev) != 2) || (portLev > 255)) // input in HEX
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			setPortLevel(FB_PORTB_, portLev);
		}
		else if (port == 'd')
		{
			setPortLevel(FB_PORTD_, portLev);
		}
		else{
			printPString("Supported ports - B and D");
			printBadArgument();
		}
	}
	else if (strstr( (char*) inputBuffer, "fbreadioportlevel") == inputBuffer) 
	{
		///read Fanboard IO port value
		char port;
		if (sscanf( (char*) inputBuffer, "fbreadioportlevel %c", &port) != 1) 
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			readPortLevel(FB_PORTB_);
		}
		else if (port == 'd')
		{
			readPortLevel(FB_PORTD_);
		}
		else{
			printPString("Supported ports - B and D");
			printBadArgument();
		}
	}	
	//=====================================================================================================================
	
	else if (strlen( (char*) inputBuffer)!=0)
	{
		printUnknownCommand();
	}
	fflush(stdout);
	printPrompt();
	fflush(stdout);
	clearBuffer();
}

void printPrompt( void )
{
	printPString("\r\n# ");
}

void printBadArgument( void )
{
	
	printPString("Bad argument(s) ");
	printf( (char*) inputBuffer);
}

void printUnknownCommand( void )
{
	printPString("Unknown command");
	#if 0
	printf( (char*) inputBuffer);
	#endif
}

void printChannelHeader( void )
{
	printPString("chan state mV mA name");
}


char uppercase(char c)
{
	if ((c>='a') && (c<'z')) return c-('a'-'A');
	return c;
}
	

/**
 * \brief Prints channel info
 * 
 * \param nr Channel number
 * 
 * \return void
 */
void printChannel( int nr )
{
	uint8_t status = 0;
	char name_upper[EEPROM_CHNAME_LEN + 1];
	//Uppercase ChannelName
	for (int i=0; i<strlen(channels[nr].name);i++) name_upper[i]=uppercase(channels[nr].name[i]);
	if (channels[nr].actual_state) status |= STATUS_ACTUAL_STATE;
	if (channels[nr].i2c_error) status |= STATUS_I2C_ERROR;
	if (channels[nr].alarm_pin0_status) status |= STATUS_ALARM_0;
	if (channels[nr].alarm_pin1_status) status |= STATUS_ALARM_1;
	if (channels[nr].overcurrent_alarm_status) status |= STATUS_OVERCURRENT;
	if (channels[nr].actual_state != channels[nr].on_off_state) status |= STATUS_UNEXPECTED;
	int voltage = (((channels[nr].negative) ? -1 : 1) * channels[nr].voltage);
	printf("\r\n%1d %02X %6d %5"PRIu32" %-20s", nr,
	status,
	voltage,
	channels[nr].current, 
	name_upper);
}

void printAlarmMasks( void )
{
	printf("%02X ", getAlarmMask());
}

void printPowerStateWord( void )
{
	uint8_t word = readRegister(CPLD_ADDR_RELAY_STATUS);
	/*for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		word |= ((channels[i].state?1:0) << i);
	}*/
	printf("%02X", word);
}

void printLimits( void )
{
	for (int channel = 0; channel < CHANNEL_COUNT; ++channel)
	{
		printf("%1d %-5"PRIu32"\r\n", channel, channels[channel].current_limit);
	}
}

void printWelcomeMessage( void )
{
	printPString("Fuse Relay Board\r\n");
	printPrompt();
}

void printPinAlarmMasks( void )
{
	uint8_t pin0 = readRegister(CPLD_ADDR_ALARMPIN0_MASK);
	uint8_t pin1 = readRegister(CPLD_ADDR_ALARMPIN1_MASK);
	printf("%02X %02X", pin0, pin1);
}

void printVersion( void )
{
	printPString("Firmware: ");
	uint8_t cpldversion=readRegister(CPLD_ADDR_VERSION);
	printf("%02d.%02d CPLD 0x%02x (%d)\r\n", 
		SVN_REVISION_MAJOR, SVN_REVISION_MINOR,  
		cpldversion, cpldversion);
	uint8_t addr[6] = {0};
	if (i2c_read_uuid(addr))
	{
		printf("UUID: %02X-%02X-%02X-%02X-%02X-%02X\r\n", 
			addr[0], addr[1], addr[2], addr[3], 
			addr[4], addr[5]);
	}
	else
	{
		printf_P(PROGMEM_STRING("UUID: 00-00-00-00-00-00\r\n"));
	}
	char box_name[64];
	eeprom_read_box_name(box_name);
	printf("Device name: %s\r\n", box_name);
	eeprom_read_inv_no(box_name);
	printf("Inventory number: %s\r\n", box_name);
	printPString("Board revision: ");
	fflush(stdout);
	switch(get_hwrevision())
	{
		case FRED2A:
			printPString("FRED2 Rev. A or FRED1 Rev. A");
			break;
		case FRED2B:
			printPString("FRED2 Rev. B");
			break;
        case FRED2C:
            printPString("FRED2 Rev. C");
            break;   
		case FRED3MC:
			printPString("FRED3M Rev. C");
			break;
		case FRED3A:
		    printPString("FRED3 Rev. A");
		    break;
        case FRED3B:
            printPString("FRED3 Rev. B");
        break;
        case FRED3MD:
            printPString("FRED3M Rev. D");
        break;
        
        default:
			printPString("UNKNOWN");
			break;
	}
	printPString("\r\n");
	printf("Protocol version: %d.%d.%d\n", PROTOCOL_VER_MAJOR, PROTOCOL_VER_MINOR, PROTOCOL_VER_REV);
	fflush(stdout);
}

/**
 * \brief Print command list
 * 
 * \return void
 */
void printHelp()
{
	printf_P(PROGMEM_STRING("readChan chanNr\r\n"
	"readAll\r\n"
	"setPwrStates\r\n"
	"chanOn chanNr\r\n"
	"chanOff chanNr\r\n"
	"setChanName channel name\r\n"
	"readCurLimits\r\n"
	"progCurLimit chanNr currentLimit\r\n"
	"powerOn\r\n"
	"powerOff\r\n"
	"readAlarms\r\n"
	"clearAlarms\r\n"
	"setAlarmMasks overcurrentAlarmMask pin0AlarmMask pin1AlarmMask\r\n"
	"readAlarmMasks\r\n"
	"setPwrMasks unconditional hotplug\r\n"
	"readPwrMasks"
	"\r\n"
	"settmcbpresence j-HWjumper p-present a-absebt\r\n"
	"TmcbSelectProm value\r\n"
	"TmcbReloadFPGA\r\n"
	"TmcbReadFPGAstatus\r\n"
	"TmcbReadPinVoltages\r\n"
	"TmcbHardReset\r\n"
	"\r\n"
	"FbSetShunt\r\n"
	"FbProgFan mode speed\r\n"
	"FbReadFansInfo\r\n"
	"FbSetIOportDir port value\r\n"
	"FbSetIOportLevel port value\r\n"
	"FbReadIOportLevel\r\n"
	"\r\n"
	"help\r\n"
	"progDeviceName name\r\n"
	"progInvNr invNumber\r\n"
	"version\r\n"
	"printstats\r\n"
	"fwupgrade\r\n"
    "reset\r\n"
	"reseti2c\r\n"));
}
