﻿/**
 * @file usart_lib.C
 * @brief USART support library
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Piotr Perek (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <usart.h>
#include <usart_serial.h>
#include <conf_usart_serial.h>
#include <asf.h>
#include <ctype.h>
#include "usart_lib.h"
#include "config/conf_usart_serial.h"
#include <stdio.h>
#include "telnet_protocol.h"
#include <progmem.h>
#include <util/delay.h>

volatile bool usartEcho = false;
static USART_Buffer_t usart_buffer;
static FILE uartStream = FDEV_SETUP_STREAM(USART_TXBuffer_PutByte, NULL, _FDEV_SETUP_WRITE);

/**
 * \brief Initialize UART
 * 
 */
void usart_init(void)
{
	//cli();
	//sysclk_init();
	static usart_serial_options_t usart_options = {
		.baudrate = USART_SERIAL_BAUDRATE,
		.charlength = USART_SERIAL_CHAR_LENGTH,
		.paritytype = USART_SERIAL_PARITY,
		.stopbits = USART_SERIAL_STOP_BIT
	};

	usart_serial_init(USART_SERIAL, &usart_options);
	usart_set_rx_interrupt_level(USART_SERIAL, PMIC_LOLVLEN_bm);
	usart_rx_enable(USART_SERIAL);
	usart_tx_enable(USART_SERIAL);
	stdout = &uartStream;
}

///USART TX empty interrupt
ISR( USARTF0_DRE_vect )
{
	USART_Buffer_t * bufPtr;
	bufPtr = &usart_buffer;

	/* Check if all data is transmitted. */
	uint16_t tempTX_Tail = usart_buffer.TX_Tail;
	if (bufPtr->TX_Head == tempTX_Tail){
		/* Disable DRE interrupts. */
		uint8_t tempCTRLA = (USART_SERIAL)->CTRLA;
		tempCTRLA = (tempCTRLA & ~USART_DREINTLVL_gm) | USART_DREINTLVL_OFF_gc;
		(USART_SERIAL)->CTRLA = tempCTRLA;

		}else{
		/* Start transmitting. */
		uint8_t data = bufPtr->TX[usart_buffer.TX_Tail];
		(USART_SERIAL)->DATA = data;
		//(USART_MMC)->DATA = data;

		/* Advance buffer tail. */
		bufPtr->TX_Tail = (bufPtr->TX_Tail + 1) & USART_TX_BUFFER_MASK;
	}
}

/**
 * \brief UART receive interrupt
 */
ISR (USARTF0_RXC_vect)
{
	ioport_toggle_pin(MMC_LED);
	uint8_t ch;
	usart_serial_getchar((USART_SERIAL), &ch);
	if (usartEcho)
	{	
		usart_serial_putchar((USART_SERIAL), ch);
		if (ch == '\r')	usart_serial_putchar((USART_SERIAL), '\n');
	}
	telnetPutChar(ch);
}

bool USART_TXBuffer_FreeSpace(void)
{
	/* Make copies to make sure that volatile access is specified. */
	uint16_t tempHead = (usart_buffer.TX_Head + 1) & USART_TX_BUFFER_MASK;
	uint16_t tempTail = usart_buffer.TX_Tail;

	/* There are data left in the buffer unless Head and Tail are equal. */
	return (tempHead != tempTail);
}

int USART_TXBuffer_PutByte(char data, FILE *stream)
{
	uint8_t tempCTRLA;
	uint16_t tempTX_Head;
	bool TXBuffer_FreeSpace;
	USART_Buffer_t * TXbufPtr;

	TXbufPtr = &usart_buffer;
	TXBuffer_FreeSpace = USART_TXBuffer_FreeSpace();

	if(TXBuffer_FreeSpace)
	{
		tempTX_Head = TXbufPtr->TX_Head;
		TXbufPtr->TX[tempTX_Head]= data;
		/* Advance buffer head. */
		TXbufPtr->TX_Head = (tempTX_Head + 1) & USART_TX_BUFFER_MASK;

		/* Enable DRE interrupt. */
		tempCTRLA = (USART_SERIAL)->CTRLA;
		tempCTRLA = (tempCTRLA & ~USART_DREINTLVL_gm) | USART_DREINTLVL_LO_gc;
		(USART_SERIAL)->CTRLA = tempCTRLA;
	}
	return TXBuffer_FreeSpace;
}

/**
 * \brief Print 0-terminated string
 * 
 * \param string
 * 
 * \return void
 */
void printString(const char *string )
{
	while (*string)
	{
		usart_serial_putchar(USART_SERIAL, *(string++));
//		_delay_us(1000);
	}
}

void printProgmemString(const char * string)
{
	char ch;
	while ((ch = PROGMEM_READ_BYTE(string++)) != 0)
		usart_serial_putchar(USART_SERIAL, ch);


}
